@@ -1,11 +1,16 @@
+"""
+This a Python implementation of the next generation GPU API.
+"""
+
 from ._coreutils import help  # noqa: F401
 from .flags import *  # noqa: F401,F403
 from .enums import *  # noqa: F401,F403
 from .base import *  # noqa: F401,F403
+from .gui import WgpuCanvasInterface  # noqa: F401,F403
 from . import base
 
 
-__version__ = "0.1.6"
+__version__ = "0.2.0"
 version_info = tuple(map(int, __version__.split(".")))
 
 
@@ -22,5 +27,3 @@ def _register_backend(func, func_async):
         raise RuntimeError("WGPU backend can only be set once.")
     globals()["request_adapter"] = func
     globals()["request_adapter_async"] = func_async
-
-    # todo: auto-select upon using request_adapter?
