@@ -39,7 +39,7 @@ from weakref import WeakKeyDictionary
 
 from cffi import FFI, __version_info__ as cffi_version_info
 
-from .. import base
+from .. import base, flags
 from .. import _register_backend
 from .._coreutils import get_resource_filename
 from .._mappings import cstructfield2enum, enummap
@@ -76,29 +76,30 @@ def _get_wgpu_lib_path():
     """ Get the path to the wgpu library, taking into account the
     WGPU_LIB_PATH environment variable.
     """
-    paths = []
 
+    # If path is given, use that or fail trying
     override_path = os.getenv("WGPU_LIB_PATH", "").strip()
     if override_path:
-        paths.append(override_path)
+        return override_path
 
-    lib_filename = None
+    # Get lib filename for supported platforms
     if sys.platform.startswith("win"):  # no-cover
         lib_filename = "wgpu_native.dll"
     elif sys.platform.startswith("darwin"):  # no-cover
         lib_filename = "libwgpu_native.dylib"
     elif sys.platform.startswith("linux"):  # no-cover
         lib_filename = "libwgpu_native.so"
-    if lib_filename:
-        # Note that this can be a false positive, e.g. ARM linux.
-        embedded_path = get_resource_filename(lib_filename)
-        paths.append(embedded_path)
-
-    for path in paths:
-        if os.path.isfile(path):
-            return path
     else:  # no-cover
-        raise RuntimeError(f"Could not find WGPU library, checked: {paths}")
+        raise RuntimeError(
+            f"No WGPU library shipped for platform {sys.platform}. Set WGPU_LIB_PATH instead."
+        )
+
+    # Note that this can be a false positive, e.g. ARM linux.
+    embedded_path = get_resource_filename(lib_filename)
+    if not os.path.isfile(embedded_path):  # no-cover
+        raise RuntimeError(f"Could not find WGPU library in {embedded_path}")
+    else:
+        return embedded_path
 
 
 # Configure cffi and load the dynamic library
@@ -273,17 +274,33 @@ def _loadop_and_clear_from_value(value):
 
 
 # wgpu.help('RequestAdapterOptions', 'requestadapter', dev=True)
-def request_adapter(*, power_preference: "GPUPowerPreference"):
-    """ Request an GPUAdapter, the object that represents the implementation of WGPU.
-    This function uses the Rust WGPU library.
+def request_adapter(*, canvas, power_preference: "GPUPowerPreference"):
+    """ Get a :class:`GPUAdapter`, the object that represents an abstract wgpu
+    implementation, from which one can request a :class:`GPUDevice`.
 
-    Params:
-        power_preference(enum): "high-performance" or "low-power"
+    This is the implementation based on the Rust wgpu-native library.
+
+    Arguments:
+        canvas (WgpuCanvas): The canvas that the adapter should be able to
+            render to (to create a swap chain for, to be precise). Can be None
+            if you're not rendering to screen (or if you're confident that the
+            returned adapter will work just fine).
+        powerPreference(PowerPreference): "high-performance" or "low-power"
     """
 
+    # Get surface id that the adapter must be compatible with. If we
+    # don't pass a valid surface id, there is no guarantee we'll be
+    # able to create a swapchain for it (from this adapter).
+    if canvas is None:
+        surface_id = 0
+    else:
+        surface_id = get_surface_id_from_canvas(canvas)
+
     # Convert the descriptor
     struct = new_struct(
-        "WGPURequestAdapterOptions *", power_preference=power_preference
+        "WGPURequestAdapterOptions *",
+        power_preference=power_preference,
+        compatible_surface=surface_id,
     )
 
     # Select possible backends. This is not exposed in the WebGPU API
@@ -299,7 +316,7 @@ def request_adapter(*, power_preference: "GPUPowerPreference"):
 
     adapter_id = None
 
-    @ffi.callback("void(uint64_t, void *)")
+    @ffi.callback("void(unsigned long long, void *)")
     def _request_adapter_callback(received, userdata):
         nonlocal adapter_id
         adapter_id = received
@@ -317,11 +334,11 @@ def request_adapter(*, power_preference: "GPUPowerPreference"):
 
 
 # wgpu.help('RequestAdapterOptions', 'requestadapter', dev=True)
-async def request_adapter_async(*, power_preference: "GPUPowerPreference"):
+async def request_adapter_async(*, canvas, power_preference: "GPUPowerPreference"):
     """ Async version of ``request_adapter()``.
     This function uses the Rust WGPU library.
     """
-    return request_adapter(power_preference=power_preference)  # no-cover
+    return request_adapter(canvas=canvas, power_preference=power_preference)  # no-cover
 
 
 # Mark as the backend at import time
@@ -342,13 +359,9 @@ class GPUAdapter(base.GPUAdapter):
         limits: "GPULimits" = {},
     ):
 
-        # Fill in defaults of limits
-        limits = limits or {}
-        limits2 = {}
-        for key in ["max_bind_groups"]:
-            limits2[key] = limits.get(key, base.default_limits[key])
-
-        extensions = tuple(extensions)
+        # Handle default limits
+        limits2 = base.default_limits.copy()
+        limits2.update(limits or {})
 
         c_extensions = new_struct(
             "WGPUExtensions *",
@@ -361,12 +374,18 @@ class GPUAdapter(base.GPUAdapter):
             "WGPUDeviceDescriptor *", extensions=c_extensions[0], limits=c_limits[0]
         )
 
-        id = _lib.wgpu_adapter_request_device(self._id, struct)
+        device_id = _lib.wgpu_adapter_request_device(self._id, struct)
 
-        queue_id = _lib.wgpu_device_get_queue(id)
+        # Get the actual limits reported by the device
+        c_limits = new_struct("WGPULimits *")
+        _lib.wgpu_device_get_limits(device_id, c_limits)
+        limits3 = {key: getattr(c_limits, key) for key in dir(c_limits)}
+
+        # Get the queue to which commands can be submitted
+        queue_id = _lib.wgpu_device_get_default_queue(device_id)
         queue = GPUQueue("", queue_id, self)
 
-        return GPUDevice(label, id, self, extensions, limits2, queue)
+        return GPUDevice(label, device_id, self, extensions, limits3, queue)
 
     # wgpu.help('DeviceDescriptor', 'adapterrequestdevice', dev=True)
     async def request_device_async(
@@ -376,31 +395,36 @@ class GPUAdapter(base.GPUAdapter):
         extensions: "GPUExtensionName-list" = [],
         limits: "GPULimits" = {},
     ):
-        return self.request_device(
-            label=label, extensions=extensions, limits=limits
-        )  # no-cover
+        return self.request_device(extensions=extensions, limits=limits)  # no-cover
+
+    def _destroy(self):
+        if self._id is not None:
+            self._id, id = None, self._id
+            _lib.wgpu_adapter_destroy(id)
 
 
 class GPUDevice(base.GPUDevice):
     # wgpu.help('BufferDescriptor', 'devicecreatebuffer', dev=True)
-    def create_buffer(
-        self, *, label="", size: "GPUSize64", usage: "GPUBufferUsageFlags"
-    ):
+    def create_buffer(self, *, label="", size: int, usage: "GPUBufferUsageFlags"):
+        c_label = ffi.new("char []", label.encode())
         size = int(size)
-
-        struct = new_struct("WGPUBufferDescriptor *", size=size, usage=usage)
+        struct = new_struct(
+            "WGPUBufferDescriptor *", label=c_label, size=size, usage=usage
+        )
 
         id = _lib.wgpu_device_create_buffer(self._internal, struct)
         return GPUBuffer(label, id, self, size, usage, "unmapped", None)
 
     # wgpu.help('BufferDescriptor', 'devicecreatebuffermapped', dev=True)
     def create_buffer_mapped(
-        self, *, label="", size: "GPUSize64", usage: "GPUBufferUsageFlags"
+        self, *, label="", size: int, usage: "GPUBufferUsageFlags"
     ):
 
+        c_label = ffi.new("char []", label.encode())
         size = int(size)
-
-        struct = new_struct("WGPUBufferDescriptor *", size=size, usage=usage)
+        struct = new_struct(
+            "WGPUBufferDescriptor *", label=c_label, size=size, usage=usage
+        )
 
         # Pointer that device_create_buffer_mapped sets, so that we can write stuff
         # there
@@ -422,21 +446,22 @@ class GPUDevice(base.GPUDevice):
         *,
         label="",
         size: "GPUExtent3D",
-        array_layer_count: "GPUIntegerCoordinate" = 1,
         mip_level_count: "GPUIntegerCoordinate" = 1,
         sample_count: "GPUSize32" = 1,
         dimension: "GPUTextureDimension" = "2d",
         format: "GPUTextureFormat",
         usage: "GPUTextureUsageFlags",
     ):
+        c_label = ffi.new("char []", label.encode())
         size = _tuple_from_tuple_or_dict(size, ("width", "height", "depth"))
         c_size = new_struct(
             "WGPUExtent3d *", width=size[0], height=size[1], depth=size[2],
         )
         struct = new_struct(
             "WGPUTextureDescriptor *",
+            label=c_label,
             size=c_size[0],
-            array_layer_count=array_layer_count,
+            array_layer_count=1,
             mip_level_count=mip_level_count,
             sample_count=sample_count,
             dimension=dimension,
@@ -460,7 +485,7 @@ class GPUDevice(base.GPUDevice):
         mipmap_filter: "GPUFilterMode" = "nearest",
         lod_min_clamp: float = 0,
         lod_max_clamp: float = 0xFFFFFFFF,
-        compare: "GPUCompareFunction" = "never",
+        compare: "GPUCompareFunction" = None,
     ):
         struct = new_struct(
             "WGPUSamplerDescriptor *",
@@ -471,7 +496,7 @@ class GPUDevice(base.GPUDevice):
             mipmap_filter=mipmap_filter,
             lod_min_clamp=lod_min_clamp,
             lod_max_clamp=lod_max_clamp,
-            compare_function=compare,
+            compare=0 if compare is None else compare,
         )
 
         id = _lib.wgpu_device_create_sampler(self._internal, struct)
@@ -479,33 +504,43 @@ class GPUDevice(base.GPUDevice):
 
     # wgpu.help('BindGroupLayoutDescriptor', 'devicecreatebindgrouplayout', dev=True)
     def create_bind_group_layout(
-        self, *, label="", bindings: "GPUBindGroupLayoutBinding-list"
+        self, *, label="", entries: "GPUBindGroupLayoutEntry-list"
     ):
-
-        c_bindings_list = []
-        for binding in bindings:
-            c_binding = new_struct(
-                "WGPUBindGroupLayoutBinding *",
-                binding=int(binding["binding"]),
-                visibility=int(binding["visibility"]),
-                ty=binding["type"],
-                texture_dimension=binding.get("texture_dimension", "2d"),
-                # ???=binding.get("textureComponentType", "float"),
-                multisampled=bool(binding.get("multisampled", False)),
-                dynamic=bool(binding.get("has_dynamic_offset", False)),
+        c_entries_list = []
+        for entry in entries:
+            type = entry["type"]
+            if "texture" in type:
+                need = {"view_dimension", "texture_component_type"}
+                if "storage" in type:
+                    need.add("storage_texture_format")
+                assert all(
+                    x in entry for x in need
+                ), f"{type} binding should specify {need}"
+            c_entry = new_struct(
+                "WGPUBindGroupLayoutEntry *",
+                binding=int(entry["binding"]),
+                visibility=int(entry["visibility"]),
+                ty=type,
+                multisampled=bool(entry.get("multisampled", False)),
+                has_dynamic_offset=bool(entry.get("has_dynamic_offset", False)),
+                view_dimension=entry.get("view_dimension", "2d"),
+                texture_component_type=entry.get("texture_component_type", "float"),
+                storage_texture_format=entry.get("storage_texture_format", 0),
             )
-            c_bindings_list.append(c_binding[0])
+            c_entries_list.append(c_entry[0])
 
-        c_bindings_array = ffi.new("WGPUBindGroupLayoutBinding []", c_bindings_list)
+        c_label = ffi.new("char []", label.encode())
+        c_entries_array = ffi.new("WGPUBindGroupLayoutEntry []", c_entries_list)
         struct = new_struct(
             "WGPUBindGroupLayoutDescriptor *",
-            bindings=c_bindings_array,
-            bindings_length=len(c_bindings_list),
+            label=c_label,
+            entries=c_entries_array,
+            entries_length=len(c_entries_list),
         )
 
         id = _lib.wgpu_device_create_bind_group_layout(self._internal, struct)
 
-        return base.GPUBindGroupLayout(label, id, self, bindings)
+        return base.GPUBindGroupLayout(label, id, self, entries)
 
     # wgpu.help('BindGroupDescriptor', 'devicecreatebindgroup', dev=True)
     def create_bind_group(
@@ -513,13 +548,13 @@ class GPUDevice(base.GPUDevice):
         *,
         label="",
         layout: "GPUBindGroupLayout",
-        bindings: "GPUBindGroupBinding-list",
+        entries: "GPUBindGroupEntry-list",
     ):
 
-        c_bindings_list = []
-        for binding in bindings:
+        c_entries_list = []
+        for entry in entries:
             # The resource can be a sampler, texture view, or buffer descriptor
-            resource = binding["resource"]
+            resource = entry["resource"]
             if isinstance(resource, base.GPUSampler):
                 c_resource_kwargs = {
                     "tag": 1,  # WGPUBindingResource_Tag.WGPUBindingResource_Sampler
@@ -536,7 +571,7 @@ class GPUDevice(base.GPUDevice):
                     )[0],
                 }
             elif isinstance(resource, dict):  # Buffer binding
-                c_buffer_binding = new_struct(
+                c_buffer_entry = new_struct(
                     "WGPUBufferBinding *",
                     buffer=resource["buffer"]._internal,
                     offset=resource["offset"],
@@ -545,29 +580,31 @@ class GPUDevice(base.GPUDevice):
                 c_resource_kwargs = {
                     "tag": 0,  # WGPUBindingResource_Tag.WGPUBindingResource_Buffer
                     "buffer": new_struct(
-                        "WGPUBindingResource_WGPUBuffer_Body *", _0=c_buffer_binding[0]
+                        "WGPUBindingResource_WGPUBuffer_Body *", _0=c_buffer_entry[0]
                     )[0],
                 }
             else:
-                raise TypeError("Unexpected resource type {type(resource)}")  # no-cover
+                raise TypeError(f"Unexpected resource type {type(resource)}")
             c_resource = new_struct("WGPUBindingResource *", **c_resource_kwargs)
-            c_binding = new_struct(
-                "WGPUBindGroupBinding *",
-                binding=int(binding["binding"]),
+            c_entry = new_struct(
+                "WGPUBindGroupEntry *",
+                binding=int(entry["binding"]),
                 resource=c_resource[0],
             )
-            c_bindings_list.append(c_binding[0])
+            c_entries_list.append(c_entry[0])
 
-        c_bindings_array = ffi.new("WGPUBindGroupBinding []", c_bindings_list)
+        c_label = ffi.new("char []", label.encode())
+        c_entries_array = ffi.new("WGPUBindGroupEntry []", c_entries_list)
         struct = new_struct(
             "WGPUBindGroupDescriptor *",
+            label=c_label,
             layout=layout._internal,
-            bindings=c_bindings_array,
-            bindings_length=len(c_bindings_list),
+            entries=c_entries_array,
+            entries_length=len(c_entries_list),
         )
 
         id = _lib.wgpu_device_create_bind_group(self._internal, struct)
-        return base.GPUBindGroup(label, id, self, bindings)
+        return base.GPUBindGroup(label, id, self, entries)
 
     # wgpu.help('PipelineLayoutDescriptor', 'devicecreatepipelinelayout', dev=True)
     def create_pipeline_layout(
@@ -587,7 +624,7 @@ class GPUDevice(base.GPUDevice):
         return base.GPUPipelineLayout(label, id, self, bind_group_layouts)
 
     # wgpu.help('ShaderModuleDescriptor', 'devicecreateshadermodule', dev=True)
-    def create_shader_module(self, *, label="", code: "GPUShaderCode"):
+    def create_shader_module(self, *, label="", code: str):
 
         if isinstance(code, bytes):
             data = code  # Assume it's Spirv
@@ -645,11 +682,11 @@ class GPUDevice(base.GPUDevice):
         label="",
         layout: "GPUPipelineLayout",
         vertex_stage: "GPUProgrammableStageDescriptor",
-        fragment_stage: "GPUProgrammableStageDescriptor",
+        fragment_stage: "GPUProgrammableStageDescriptor" = None,
         primitive_topology: "GPUPrimitiveTopology",
         rasterization_state: "GPURasterizationStateDescriptor" = {},
         color_states: "GPUColorStateDescriptor-list",
-        depth_stencil_state: "GPUDepthStencilStateDescriptor",
+        depth_stencil_state: "GPUDepthStencilStateDescriptor" = None,
         vertex_state: "GPUVertexStateDescriptor" = {},
         sample_count: "GPUSize32" = 1,
         sample_mask: "GPUSampleMask" = 0xFFFFFFFF,
@@ -660,18 +697,20 @@ class GPUDevice(base.GPUDevice):
             module=vertex_stage["module"]._internal,
             entry_point=ffi.new("char []", vertex_stage["entry_point"].encode()),
         )
-        c_fragment_stage = new_struct(
-            "WGPUProgrammableStageDescriptor *",
-            module=fragment_stage["module"]._internal,
-            entry_point=ffi.new("char []", fragment_stage["entry_point"].encode()),
-        )
+        c_fragment_stage = ffi.NULL
+        if fragment_stage is not None:
+            c_fragment_stage = new_struct(
+                "WGPUProgrammableStageDescriptor *",
+                module=fragment_stage["module"]._internal,
+                entry_point=ffi.new("char []", fragment_stage["entry_point"].encode()),
+            )
         c_rasterization_state = new_struct(
             "WGPURasterizationStateDescriptor *",
-            front_face=rasterization_state["front_face"],
-            cull_mode=rasterization_state["cull_mode"],
-            depth_bias=rasterization_state["depth_bias"],
-            depth_bias_slope_scale=rasterization_state["depth_bias_slope_scale"],
-            depth_bias_clamp=rasterization_state["depth_bias_clamp"],
+            front_face=rasterization_state.get("front_face", "ccw"),
+            cull_mode=rasterization_state.get("cull_mode", "none"),
+            depth_bias=rasterization_state.get("depth_bias", 0),
+            depth_bias_slope_scale=rasterization_state.get("depth_bias_slope_scale", 0),
+            depth_bias_clamp=rasterization_state.get("depth_bias_clamp", 0),
         )
         c_color_states_list = []
         for color_state in color_states:
@@ -698,8 +737,8 @@ class GPUDevice(base.GPUDevice):
                 format=color_state["format"],
                 alpha_blend=c_alpha_blend[0],
                 color_blend=c_color_blend[0],
-                write_mask=color_state["write_mask"],
-            )  # enum
+                write_mask=color_state.get("write_mask", 0xF),
+            )
             c_color_states_list.append(c_color_state[0])
         c_color_states_array = ffi.new(
             "WGPUColorStateDescriptor []", c_color_states_list
@@ -707,31 +746,37 @@ class GPUDevice(base.GPUDevice):
         if depth_stencil_state is None:
             c_depth_stencil_state = ffi.NULL
         else:
-            stencil_front = depth_stencil_state["stencil_front"]
+            stencil_front = depth_stencil_state.get("stencil_front", {})
             c_stencil_front = new_struct(
                 "WGPUStencilStateFaceDescriptor *",
-                compare=stencil_front["compare"],
-                fail_op=stencil_front["fail_op"],
-                depth_fail_op=stencil_front["depth_fail_op"],
-                pass_op=stencil_front["pass_op"],
+                compare=stencil_front.get("compare", "always"),
+                fail_op=stencil_front.get("fail_op", "keep"),
+                depth_fail_op=stencil_front.get("depth_fail_op", "keep"),
+                pass_op=stencil_front.get("pass_op", "keep"),
             )
-            stencil_back = depth_stencil_state["stencil_back"]
+            stencil_back = depth_stencil_state.get("stencil_back", {})
             c_stencil_back = new_struct(
                 "WGPUStencilStateFaceDescriptor *",
-                compare=stencil_back["compare"],
-                fail_op=stencil_back["fail_op"],
-                depth_fail_op=stencil_back["depth_fail_op"],
-                pass_op=stencil_back["pass_op"],
+                compare=stencil_back.get("compare", "always"),
+                fail_op=stencil_back.get("fail_op", "keep"),
+                depth_fail_op=stencil_back.get("depth_fail_op", "keep"),
+                pass_op=stencil_back.get("pass_op", "keep"),
             )
             c_depth_stencil_state = new_struct(
                 "WGPUDepthStencilStateDescriptor *",
                 format=depth_stencil_state["format"],
-                depth_write_enabled=bool(depth_stencil_state["depth_write_enabled"]),
-                depth_compare=depth_stencil_state["depth_compare"],
+                depth_write_enabled=bool(
+                    depth_stencil_state.get("depth_write_enabled", False)
+                ),
+                depth_compare=depth_stencil_state.get("depth_compare", "always"),
                 stencil_front=c_stencil_front[0],
                 stencil_back=c_stencil_back[0],
-                stencil_read_mask=depth_stencil_state["stencil_read_mask"],
-                stencil_write_mask=depth_stencil_state["stencil_write_mask"],
+                stencil_read_mask=depth_stencil_state.get(
+                    "stencil_read_mask", 0xFFFFFFFF
+                ),
+                stencil_write_mask=depth_stencil_state.get(
+                    "stencil_write_mask", 0xFFFFFFFF
+                ),
             )
         c_vertex_buffer_descriptors_list = []
         for buffer_des in vertex_state["vertex_buffers"]:
@@ -748,19 +793,19 @@ class GPUDevice(base.GPUDevice):
                 "WGPUVertexAttributeDescriptor []", c_attributes_list
             )
             c_vertex_buffer_descriptor = new_struct(
-                "WGPUVertexBufferDescriptor *",
-                stride=buffer_des["array_stride"],
-                step_mode=buffer_des["stepmode"],
+                "WGPUVertexBufferLayoutDescriptor *",
+                array_stride=buffer_des["array_stride"],
+                step_mode=buffer_des.get("step_mode", "vertex"),
                 attributes=c_attributes_array,
                 attributes_length=len(c_attributes_list),
             )
             c_vertex_buffer_descriptors_list.append(c_vertex_buffer_descriptor[0])
         c_vertex_buffer_descriptors_array = ffi.new(
-            "WGPUVertexBufferDescriptor []", c_vertex_buffer_descriptors_list
+            "WGPUVertexBufferLayoutDescriptor []", c_vertex_buffer_descriptors_list
         )
-        c_vertex_input = new_struct(
-            "WGPUVertexInputDescriptor *",
-            index_format=vertex_state["index_format"],
+        c_vertex_state = new_struct(
+            "WGPUVertexStateDescriptor *",
+            index_format=vertex_state.get("index_format", "uint32"),
             vertex_buffers=c_vertex_buffer_descriptors_array,
             vertex_buffers_length=len(c_vertex_buffer_descriptors_list),
         )
@@ -775,7 +820,7 @@ class GPUDevice(base.GPUDevice):
             color_states=c_color_states_array,
             color_states_length=len(c_color_states_list),
             depth_stencil_state=c_depth_stencil_state,
-            vertex_input=c_vertex_input[0],
+            vertex_state=c_vertex_state[0],
             sample_count=sample_count,
             sample_mask=sample_mask,
             alpha_to_coverage_enabled=alpha_to_coverage_enabled,
@@ -786,8 +831,8 @@ class GPUDevice(base.GPUDevice):
 
     # wgpu.help('CommandEncoderDescriptor', 'devicecreatecommandencoder', dev=True)
     def create_command_encoder(self, *, label=""):
-
-        struct = new_struct("WGPUCommandEncoderDescriptor *", todo=0)
+        c_label = ffi.new("char []", label.encode())
+        struct = new_struct("WGPUCommandEncoderDescriptor *", label=c_label)
 
         id = _lib.wgpu_device_create_command_encoder(self._internal, struct)
         return GPUCommandEncoder(label, id, self)
@@ -798,15 +843,13 @@ class GPUDevice(base.GPUDevice):
     #     *,
     #     label="",
     #     color_formats: "GPUTextureFormat-list",
-    #     depth_stencil_format: "GPUTextureFormat",
+    #     depth_stencil_format: "GPUTextureFormat" = None,
     #     sample_count: "GPUSize32" = 1,
     # ):
     #     pass
 
-    def _gui_configure_swap_chain(self, canvas, format, usage):
-        """ Get a swapchain object from a canvas object. Called by BaseCanvas.
-        """
-        # Note: canvas should implement the BaseCanvas interface.
+    def configure_swap_chain(self, canvas, format, usage=None):
+        usage = flags.TextureUsage.OUTPUT_ATTACHMENT if usage is None else usage
         return GPUSwapChain(self, canvas, format, usage)
 
 
@@ -835,6 +878,7 @@ class GPUBuffer(base.GPUBuffer):
             raise RuntimeError("Could not read buffer data.")
 
         self._state = "mapped"
+        self._mapping = data
         return data
 
     # wgpu.help('buffermapwriteasync', dev=True)
@@ -861,6 +905,7 @@ class GPUBuffer(base.GPUBuffer):
             raise RuntimeError("Could not read buffer data.")
 
         self._state = "mapped"
+        self._mapping = data
         return data
 
     # wgpu.help('buffermapreadasync', dev=True)
@@ -878,56 +923,74 @@ class GPUBuffer(base.GPUBuffer):
         if self._state == "mapped":
             _lib.wgpu_buffer_unmap(self._internal)
             self._state = "unmapped"
+            self._mapping = None
 
     # wgpu.help('bufferdestroy', dev=True)
     def destroy(self):
-        if self._state != "destroyed":
+        self._destroy()  # no-cover
+
+    def _destroy(self):
+        if self._internal is not None:
+            self._internal, internal = None, self._internal
             self._state = "destroyed"
-            _lib.wgpu_buffer_destroy(self._internal)
+            self._mapping = None
+            _lib.wgpu_buffer_destroy(internal)
 
 
 class GPUTexture(base.GPUTexture):
 
-    _destroyed = False
-
     # wgpu.help('TextureViewDescriptor', 'texturecreateview', dev=True)
     def create_view(
         self,
         *,
         label="",
-        format: "GPUTextureFormat",
-        dimension: "GPUTextureViewDimension",
+        format: "GPUTextureFormat" = None,
+        dimension: "GPUTextureViewDimension" = None,
         aspect: "GPUTextureAspect" = "all",
         base_mip_level: "GPUIntegerCoordinate" = 0,
         mip_level_count: "GPUIntegerCoordinate" = 0,
         base_array_layer: "GPUIntegerCoordinate" = 0,
         array_layer_count: "GPUIntegerCoordinate" = 0,
     ):
+        if format is None or dimension is None:
+            if not (
+                format is None
+                and dimension is None
+                and aspect == "all"
+                and base_mip_level == 0
+                and mip_level_count == 0
+                and base_array_layer == 0
+                and array_layer_count == 0
+            ):
+                raise ValueError(
+                    "In create_view() if any paramter is given, "
+                    + "both format and dimension must be specified."
+                )
+            id = _lib.wgpu_texture_create_view(self._internal, ffi.NULL)
 
-        struct = new_struct(
-            "WGPUTextureViewDescriptor *",
-            format=format,
-            dimension=dimension,
-            aspect=aspect,
-            base_mip_level=base_mip_level,
-            level_count=mip_level_count,
-            base_array_layer=base_array_layer,
-            array_layer_count=array_layer_count,
-        )
+        else:
+            struct = new_struct(
+                "WGPUTextureViewDescriptor *",
+                format=format,
+                dimension=dimension,
+                aspect=aspect,
+                base_mip_level=base_mip_level,
+                level_count=mip_level_count,
+                base_array_layer=base_array_layer,
+                array_layer_count=array_layer_count,
+            )
+            id = _lib.wgpu_texture_create_view(self._internal, struct)
 
-        id = _lib.wgpu_texture_create_view(self._internal, struct)
-        return base.GPUTextureView(label, id, self)
-
-    def create_default_view(self, *, label=""):
-        # This method is available in wgpu-rs, and it's kinda nice :)
-        id = _lib.wgpu_texture_create_view(self._internal, ffi.NULL)
         return base.GPUTextureView(label, id, self)
 
     # wgpu.help('texturedestroy', dev=True)
     def destroy(self):
-        if not self._destroyed:
-            self._destroyed = True
-            _lib.wgpu_texture_destroy(self._internal)
+        self._destroy()  # no-cover
+
+    def _destroy(self):
+        if self._internal is not None:
+            self._internal, internal = None, self._internal
+            _lib.wgpu_texture_destroy(internal)
 
 
 class GPUCommandEncoder(base.GPUCommandEncoder):
@@ -943,16 +1006,18 @@ class GPUCommandEncoder(base.GPUCommandEncoder):
         *,
         label="",
         color_attachments: "GPURenderPassColorAttachmentDescriptor-list",
-        depth_stencil_attachment: "GPURenderPassDepthStencilAttachmentDescriptor",
+        depth_stencil_attachment: "GPURenderPassDepthStencilAttachmentDescriptor" = None,
+        occlusion_query_set: "GPUQuerySet" = None,
     ):
+        # Note that occlusion_query_set is ignored because wgpu-native does not have it.
 
         c_color_attachments_list = []
         for color_attachment in color_attachments:
             assert isinstance(color_attachment["attachment"], base.GPUTextureView)
             texture_view_id = color_attachment["attachment"]._internal
             c_resolve_target = (
-                ffi.NULL
-                if color_attachment["resolve_target"] is None
+                0
+                if color_attachment.get("resolve_target", None) is None
                 else color_attachment["resolve_target"]._internal
             )  # this is a TextureViewId or null
             c_load_op, clear_color = _loadop_and_clear_from_value(
@@ -971,7 +1036,7 @@ class GPUCommandEncoder(base.GPUCommandEncoder):
                 attachment=texture_view_id,
                 resolve_target=c_resolve_target,
                 load_op=c_load_op,
-                store_op=color_attachment["store_op"],
+                store_op=color_attachment.get("store_op", "store"),
                 clear_color=c_clear_color[0],
             )
             c_color_attachments_list.append(c_attachment[0])
@@ -1030,8 +1095,8 @@ class GPUCommandEncoder(base.GPUCommandEncoder):
             "WGPUBufferCopyView *",
             buffer=source["buffer"]._internal,
             offset=int(source.get("offset", 0)),
-            row_pitch=int(source["row_pitch"]),
-            image_height=int(source["image_height"]),
+            bytes_per_row=int(source["bytes_per_row"]),
+            rows_per_image=int(source.get("rows_per_image", 0)),
         )
 
         ori = _tuple_from_tuple_or_dict(destination["origin"], "xyz")
@@ -1070,8 +1135,8 @@ class GPUCommandEncoder(base.GPUCommandEncoder):
             "WGPUBufferCopyView *",
             buffer=destination["buffer"]._internal,
             offset=int(destination.get("offset", 0)),
-            row_pitch=int(destination["row_pitch"]),
-            image_height=int(destination["image_height"]),
+            bytes_per_row=int(destination["bytes_per_row"]),
+            rows_per_image=int(destination.get("rows_per_image", 0)),
         )
 
         size = _tuple_from_tuple_or_dict(copy_size, ("width", "height", "depth"))
@@ -1121,9 +1186,14 @@ class GPUCommandEncoder(base.GPUCommandEncoder):
         id = _lib.wgpu_command_encoder_finish(self._internal, struct)
         return base.GPUCommandBuffer(label, id, self)
 
+    # todo: these do not exist yet for command_encoder in wgpu-native
+    # def push_debug_group(self, group_label):
+    # def pop_debug_group(self):
+    # def insert_debug_marker(self, marker_label):
+
 
 class GPUProgrammablePassEncoder(base.GPUProgrammablePassEncoder):
-    # wgpu.help('BindGroup', 'Index32', 'Size64', 'programmablepassencodersetbindgroup', dev=True)
+    # wgpu.help('BindGroup', 'Index32', 'Size32', 'Size64', 'programmablepassencodersetbindgroup', dev=True)
     def set_bind_group(
         self,
         index,
@@ -1133,7 +1203,7 @@ class GPUProgrammablePassEncoder(base.GPUProgrammablePassEncoder):
         dynamic_offsets_data_length,
     ):
         offsets = list(dynamic_offsets_data)
-        c_offsets = ffi.new("WGPUBufferAddress []", offsets)
+        c_offsets = ffi.new("WGPUDynamicOffset []", offsets)
         bind_group_id = bind_group._internal
         if isinstance(self, GPUComputePassEncoder):
             _lib.wgpu_compute_pass_set_bind_group(
@@ -1144,17 +1214,28 @@ class GPUProgrammablePassEncoder(base.GPUProgrammablePassEncoder):
                 self._internal, index, bind_group_id, c_offsets, len(offsets)
             )
 
-    # # wgpu.help('programmablepassencoderpushdebuggroup', dev=True)
-    # def push_debug_group(self, group_label):
-    #     ...
-    #
-    # # wgpu.help('programmablepassencoderpopdebuggroup', dev=True)
-    # def pop_debug_group(self):
-    #     ...
-    #
-    # # wgpu.help('programmablepassencoderinsertdebugmarker', dev=True)
-    # def insert_debug_marker(self, marker_label):
-    #     ...
+    # wgpu.help('programmablepassencoderpushdebuggroup', dev=True)
+    def push_debug_group(self, group_label):
+        c_group_label = ffi.new("char []", group_label.encode())
+        if isinstance(self, GPUComputePassEncoder):
+            _lib.wgpu_compute_pass_push_debug_group(self._internal, c_group_label)
+        else:
+            _lib.wgpu_render_pass_push_debug_group(self._internal, c_group_label)
+
+    # wgpu.help('programmablepassencoderpopdebuggroup', dev=True)
+    def pop_debug_group(self):
+        if isinstance(self, GPUComputePassEncoder):
+            _lib.wgpu_compute_pass_pop_debug_group(self._internal)
+        else:
+            _lib.wgpu_render_pass_pop_debug_group(self._internal)
+
+    # wgpu.help('programmablepassencoderinsertdebugmarker', dev=True)
+    def insert_debug_marker(self, marker_label):
+        c_marker_label = ffi.new("char []", marker_label.encode())
+        if isinstance(self, GPUComputePassEncoder):
+            _lib.wgpu_compute_pass_insert_debug_marker(self._internal, c_marker_label)
+        else:
+            _lib.wgpu_render_pass_insert_debug_marker(self._internal, c_marker_label)
 
 
 class GPUComputePassEncoder(GPUProgrammablePassEncoder):
@@ -1167,7 +1248,7 @@ class GPUComputePassEncoder(GPUProgrammablePassEncoder):
         _lib.wgpu_compute_pass_set_pipeline(self._internal, pipeline_id)
 
     # wgpu.help('Size32', 'computepassencoderdispatch', dev=True)
-    def dispatch(self, x, y, z):
+    def dispatch(self, x, y=1, z=1):
         _lib.wgpu_compute_pass_dispatch(self._internal, x, y, z)
 
     # wgpu.help('Buffer', 'Size64', 'computepassencoderdispatchindirect', dev=True)
@@ -1181,6 +1262,13 @@ class GPUComputePassEncoder(GPUProgrammablePassEncoder):
     def end_pass(self):
         _lib.wgpu_compute_pass_end_pass(self._internal)
 
+    # todo: const uint8_t *wgpu_compute_pass_finish(WGPURawPass *pass, uintptr_t *length);
+
+    def _destroy(self):
+        if self._internal is not None:
+            self._internal, internal = None, self._internal
+            internal  # todo: crashes _lib.wgpu_compute_pass_destroy(internal)
+
 
 class GPURenderEncoderBase(GPUProgrammablePassEncoder):
     """
@@ -1192,20 +1280,19 @@ class GPURenderEncoderBase(GPUProgrammablePassEncoder):
         _lib.wgpu_render_pass_set_pipeline(self._internal, pipeline_id)
 
     # wgpu.help('Buffer', 'Size64', 'renderencoderbasesetindexbuffer', dev=True)
-    def set_index_buffer(self, buffer, offset):
-        _lib.wgpu_render_pass_set_index_buffer(self._internal, buffer._internal, offset)
+    def set_index_buffer(self, buffer, offset=0, size=0):
+        _lib.wgpu_render_pass_set_index_buffer(
+            self._internal, buffer._internal, int(offset), int(size)
+        )
 
     # wgpu.help('Buffer', 'Index32', 'Size64', 'renderencoderbasesetvertexbuffer', dev=True)
-    def set_vertex_buffer(self, slot, buffer, offset):
-        buffers, offsets = [buffer], [offset]
-        c_buffer_ids = ffi.new("WGPUBufferId []", [b._internal for b in buffers])
-        c_offsets = ffi.new("WGPUBufferAddress []", [int(i) for i in offsets])
-        _lib.wgpu_render_pass_set_vertex_buffers(
-            self._internal, slot, c_buffer_ids, c_offsets, len(buffers)
+    def set_vertex_buffer(self, slot, buffer, offset=0, size=0):
+        _lib.wgpu_render_pass_set_vertex_buffer(
+            self._internal, int(slot), buffer._internal, int(offset), int(size)
         )
 
     # wgpu.help('Size32', 'renderencoderbasedraw', dev=True)
-    def draw(self, vertex_count, instance_count, first_vertex, first_instance):
+    def draw(self, vertex_count, instance_count=1, first_vertex=0, first_instance=0):
         _lib.wgpu_render_pass_draw(
             self._internal, vertex_count, instance_count, first_vertex, first_instance
         )
@@ -1219,7 +1306,12 @@ class GPURenderEncoderBase(GPUProgrammablePassEncoder):
 
     # wgpu.help('SignedOffset32', 'Size32', 'renderencoderbasedrawindexed', dev=True)
     def draw_indexed(
-        self, index_count, instance_count, first_index, base_vertex, first_instance
+        self,
+        index_count,
+        instance_count=1,
+        first_index=0,
+        base_vertex=0,
+        first_instance=0,
     ):
         _lib.wgpu_render_pass_draw_indexed(
             self._internal,
@@ -1237,6 +1329,13 @@ class GPURenderEncoderBase(GPUProgrammablePassEncoder):
             self._internal, buffer_id, int(indirect_offset)
         )
 
+    # todo: uint8_t *wgpu_render_pass_finish(WGPURawPass *pass, uintptr_t *length);
+
+    def _destroy(self):
+        if self._internal is not None:
+            self._internal, internal = None, self._internal
+            internal  # todo: crashes _lib.wgpu_render_pass_destroy(internal)
+
 
 class GPURenderPassEncoder(GPURenderEncoderBase):
 
@@ -1341,6 +1440,7 @@ class GPUSwapChain(base.GPUSwapChain):
             height=max(1, psize[1]),
             present_mode=1,
         )
+        # present_mode -> 0: Immediate, 1: Mailbox, 2: Fifo
 
         if self._surface_id is None:
             self._surface_id = get_surface_id_from_canvas(canvas)
@@ -1349,17 +1449,14 @@ class GPUSwapChain(base.GPUSwapChain):
             self._device._internal, self._surface_id, struct
         )
 
-    def get_current_texture_view(self):
-        # todo: should we cache instances (on their id)?
-        # otherwise we have multiple instances mapping to same internal texture
+    def __enter__(self):
+        # Get the current texture view, and make sure it is presented when done
         self._create_native_swap_chain_if_needed()
         swap_chain_output = _lib.wgpu_swap_chain_get_next_texture(self._internal)
         return base.GPUTextureView("swap_chain", swap_chain_output.view_id, self)
 
-    def _gui_present(self):
-        """ Present the current texture. This is not part of the public API,
-        instead, GUI backends should call this at the right moment.
-        """
+    def __exit__(self, type, value, tb):
+        # Present the current texture
         _lib.wgpu_swap_chain_present(self._internal)
 
 
